
;; -*- indent-tabs-mode: nil -*-
;;
;; File: undo.jl
;; Version: $Revision$
;; Description: 
;;     Remembers window activity (moving, resizing, iconifying, shading,
;;     workspace switching) and lets you undo it.
;;
;;     Do you ever accidentally move a window, and then have to 
tediously
;;     replace it where it used to be?  Ever hit a "clean up this 
cluttered
;;     desktop" command, and decide that the results are worse than 
what you
;;     started with?
;;
;; Installation:
;;     Install this file, and add the following lines to your 
.sawmillrc:
;;
;; (setq load-path (cons "/directory/containing/this/file/" load-path))
;; (require 'undo)
;;
;;     And then bind the "undo" and "undo-redo" commands to two handy 
keys.
;;
;; Author:
;;     Terry Weissman <terry@weissman.org>
;;     Latest version can be found at http://www.weissman.org/sawfish






(defun undo-eval (cmd)
  (eval cmd)
;  (display-message (format nil "%S" cmd))
)

(defun undo-record (l)
  (setq undo-curcmd-list (cons l undo-curcmd-list)))

(defun undo ()
  "Undo last window move or resize"
  (interactive)
  (let ((cmd (car undo-list))
        (x (cdr undo-list)))
    (setq undo-list ())
    (setq undo-curcmd-list ())
    (undo-eval cmd)
    (undo-transfer-to-undo-list)
    (if undo-list
        (setq undo-redo-list (cons (car undo-list) undo-redo-list)))
    (setq undo-list x)
    (setq undo-curcmd-list ())))


(defun undo-redo ()
  "Redo last window move or resize"
  (interactive)
  (let ((cmd (car undo-redo-list)))
    (setq undo-redo-list (cdr undo-redo-list))
    (undo-eval cmd)))


(defun undo-record-move (w)
  (let ((oldloc (window-get w 'undo-loc))
        (newloc (window-position w)))
    (if oldloc
        (unless (equal oldloc newloc)
          (undo-record (list 'move-window-to w (car oldloc) (cdr 
oldloc)))))
    (window-put w 'undo-loc newloc)))
  
    
(defun undo-record-resize (w)
  (let ((oldsize (window-get w 'undo-size))
        (newsize (window-dimensions w)))
    (if oldsize
        (unless (equal oldsize newsize)
          (undo-record (list 'resize-window-to w (car oldsize) (cdr 
oldsize)))))
    (window-put w 'undo-size newsize)))
  
    
(defun undo-record-workspace (w)
  (undo-record (list 'select-workspace w)))


(defun undo-record-iconify (w)
  (undo-record (list 'uniconify-window w)))

(defun undo-record-uniconify (w)
  (undo-record (list 'iconify-window w)))

(defun undo-record-shade (w)
  (undo-record (list 'unshade-window w)))

(defun undo-record-unshade (w)
  (undo-record (list 'shade-window w)))


(defun undo-transfer-to-undo-list ()
  (if undo-curcmd-list
      (progn
        (setq undo-list (cons (cons 'progn undo-curcmd-list) undo-list))
        (setq undo-curcmd-list nil)
        (while (> (length undo-list) 50)
          (setq undo-list (nreverse (cdr (nreverse undo-list))))))))

  
(defun undo-transfer-if-done ()
  (unless (or undo-in-command undo-in-interactive)
    (undo-transfer-to-undo-list)))


(defun undo-pre-command ()
  (setq undo-in-command t))

(defun undo-post-command ()
  (setq undo-in-command nil)
  (undo-transfer-if-done))

(defun undo-pre-interactive ()
  (setq undo-in-interactive t))

(defun undo-post-interactive (w)
  (setq undo-in-interactive nil)
  (undo-record-resize w)                ;These two lines shouldn't be
  (undo-record-move w)                  ;necessary, but they don't 
hurt...
  (undo-transfer-if-done))


(defun undo-record-all-locs ()
  (map-windows (lambda (w)
                 (window-put w 'undo-loc (window-position w))
                 (window-put w 'undo-size (window-dimensions w)))))



;;;;; Initialize global variables.

(unless (boundp 'undo-list)
  (setq undo-list ()))

(unless (boundp 'undo-redo-list)
  (setq undo-redo-list ()))

(unless (boundp 'undo-curcmd-list)
  (setq undo-curcmd-list ()))

(unless (boundp 'undo-in-command)
  (setq undo-in-command nil))

(unless (boundp 'undo-in-interactive)
  (setq undo-in-interactive nil))



;;;;; Set hooks for operations that we want to be able to undo.


(unless (in-hook-p 'window-moved-hook undo-record-move)
  (add-hook 'window-moved-hook undo-record-move))

(unless (in-hook-p 'window-resized-hook undo-record-resize)
  (add-hook 'window-resized-hook undo-record-resize))

(unless (in-hook-p 'leave-workspace-hook undo-record-workspace)
  (add-hook 'leave-workspace-hook undo-record-workspace))

(unless (in-hook-p 'iconify-window-hook undo-record-iconify)
  (add-hook 'iconify-window-hook undo-record-iconify))

(unless (in-hook-p 'uniconify-window-hook undo-record-uniconify)
  (add-hook 'uniconify-window-hook undo-record-uniconify))

(unless (in-hook-p 'shade-window-hook undo-record-shade)
  (add-hook 'shade-window-hook undo-record-shade))

(unless (in-hook-p 'unshade-window-hook undo-record-unshade)
  (add-hook 'unshade-window-hook undo-record-unshade))



;;;;; Set hooks for operations that tend to group un-do-able operations.

(unless (in-hook-p 'pre-command-hook undo-pre-command)
  (add-hook 'pre-command-hook undo-pre-command))

(unless (in-hook-p 'post-command-hook undo-post-command)
  (add-hook 'post-command-hook undo-post-command))

(unless (in-hook-p 'before-move-hook undo-pre-interactive)
  (add-hook 'before-move-hook undo-pre-interactive))

(unless (in-hook-p 'after-move-hook undo-post-interactive)
  (add-hook 'after-move-hook undo-post-interactive))

(unless (in-hook-p 'before-resize-hook undo-pre-interactive)
  (add-hook 'before-resize-hook undo-pre-interactive))

(unless (in-hook-p 'after-resize-hook undo-post-interactive)
  (add-hook 'after-resize-hook undo-post-interactive))


(make-timer undo-record-all-locs 1 0)


; (setq log "")
; (add-hook 'window-moved-hook (lambda () (setq log (concat log "*"))))
; (add-hook 'before-move-hook (lambda () (setq log (concat log "m"))))
; (add-hook 'after-move-hook (lambda () (setq log (concat log "M"))))
; (add-hook 'pre-command-hook (lambda () (setq log (concat log "c"))))
; (add-hook 'post-command-hook (lambda () (setq log (concat log "C"))))


