;; group.jl - Make window groups
;;
;; Author : Yann Hodique <Yann.Hodique@lifl.fr>
;;
;; Modified by Scott Scriven <sawfish@toykeeper.net>
;; (mostly hook updates)

(define-structure sigma.group

    (export sigma-group-group-window
            sigma-group-release-window
            sigma-group-offset
            sigma-group-same-group-p
            sigma-group-raise-left-window
            sigma-group-raise-right-window
            sigma-group-map-other-grouped-windows
            siggroup-p
            siggroup-position
            siggroup-dimensions
            siggroup-window-list
            sigma-group-find-window
            sigma-group-rank
            sigma-group-refresh-group
            sigma-group-released-window-hook
            sigma-group-grouped-window-hook
            sigma-group-groups)

    (open rep
          rep.system
          rep.data.records
          rep.data.tables
          sawfish.wm.windows
          sawfish.wm.stacking
          sawfish.wm.state.shading
          sigma.refs)

  ;; siggroup record is composed of a position, dimensions and a list of windows
  (define-record-type :siggroup
    (siggroup pos dim wlist)
    siggroup-p
    (pos siggroup-position)
    (dim siggroup-dimensions)
    (wlist siggroup-window-list))

  ;; siggroup printer
  (define-record-discloser :siggroup (lambda (x) `(siggroup ,(siggroup-position x) ,(siggroup-dimensions x) ,(siggroup-window-list x))))

  ;; Table containing the existing groups
  (define sigma-group-groups (sigma-make-table))

  ;; Some lock variables
  (define sigma-group-refresh-lock t)
  (define sigma-group-move-lock t)

  ;;;;;;;;;;;
  ;; Hooks :
  ;;

  (define sigma-group-released-window-hook nil)
  (define sigma-group-grouped-window-hook nil)

  ;;;;;;;;;;;;;;;
  ;; Functions :
  ;;

  ;; Move and resize according to *frame* dimensions
  (define (sigma-group-move-resize-frame-window-to win x y w h)
    (let* ((dim1 (window-dimensions win))
           (dim2 (window-frame-dimensions win))
           (dw (- (car dim2) (car dim1)))
           (dh (- (cdr dim2) (cdr dim1))))
      (move-resize-window-to win x y (- w dw) (- h dh)))
    )

  ;; Return the index of a new group containing only win
  (define (sigma-group-make-new-group win)
    (let* ((pos (window-position win))
           (dim (window-frame-dimensions win))
           (group (siggroup pos dim (list win)))
           (index (sigma-table-set sigma-group-groups group)))
      (window-put win 'siggroup index)
      index))

  ;; Return a group containing win
  (define (sigma-group-find-window win)
    (sigma-table-ref sigma-group-groups (sigma-group-window-group-index win)))

  ;; Return the index of the group containing win
  (define (sigma-group-window-group-index win)
    (let ((index (window-get win 'siggroup)))
      (if index index (sigma-group-make-new-group win))))

  ;; Rank of an element in a list
  (define (sigma-group-rank elem list)
    (if (eq elem (car list))
        0
      (+ 1 (sigma-group-rank elem (cdr list)))))

  ;; Remove a window from a group at given index
  (define (sigma-group-delete-window-from-group win index)
    (let* ((old (sigma-table-ref sigma-group-groups index))
           (l (remove win (siggroup-window-list old))))
      (if (null l)
          (sigma-table-unset sigma-group-groups index)
        (sigma-table-set sigma-group-groups (siggroup (siggroup-position old) (siggroup-dimensions old) l) index)
        (sigma-group-refresh-group (car l) 'frame)
        )))

  ;; Find window's group and remove it
  (define (sigma-group-delete-window-from-groups w)
    (sigma-group-delete-window-from-group w (sigma-group-window-group-index w)))

  ;; Put window in group at given index
  (define (sigma-group-put-window-in-group win index)
    (let* ((group (sigma-table-ref sigma-group-groups index))
           (dim (siggroup-dimensions group))
           (pos (siggroup-position group)))
      (sigma-table-set sigma-group-groups (siggroup (siggroup-position group)
                               (siggroup-dimensions group)
                               (append (siggroup-window-list group) (list win))) index)
      (window-put win 'siggroup index)
      (sigma-group-move-resize-frame-window-to win (car pos) (cdr pos) (car dim) (cdr dim))
      (rebuild-frame win)
      ))

  ;; Refresh the entire group containing win according to prop. prop can be one of the symbols : frame, move, resize, sticky, shade, unshade
  (define (sigma-group-refresh-group win prop)
    (if sigma-group-refresh-lock
        (progn
          (setq sigma-group-refresh-lock nil)
          (let* ((index (sigma-group-window-group-index win))
                 (wins (siggroup-window-list (sigma-table-ref sigma-group-groups index))))
            (cond
             ((eq prop 'frame)
              (mapcar (lambda (w)
                        (rebuild-frame w)) wins)
              )
             ((or (eq prop 'move) (eq prop 'resize))
              (let ((dim (window-frame-dimensions win))
                    (pos (window-position win)))
                (mapcar (lambda (w)
                          (sigma-group-move-resize-frame-window-to w (car pos) (cdr pos) (car dim) (cdr dim))
                          (rebuild-frame w)) wins)
                (sigma-table-set sigma-group-groups (siggroup pos dim wins) index)))
             ((eq prop 'sticky)
              (mapcar (lambda (w)
                        (toggle-window-sticky w)) wins)
              )
             ((eq prop 'shade)
              (mapcar (lambda (w)
                        (shade-window w)
                        (rebuild-frame w)) wins)
              )
             ((eq prop 'unshade)
              (mapcar (lambda (w)
                        (unshade-window w)
                        (rebuild-frame w)) wins)
              )
             ))
          (setq sigma-group-refresh-lock t)
          )))

  ;; Put active window in pointer-selected group
  (define (sigma-group-group-window w win)
    (interactive)
    (let* ((index (sigma-group-window-group-index win))
           (index2 (sigma-group-window-group-index w)))
      (sigma-group-refresh-group win 'move) ;ugly hack, don't know why it's needed, but new groups are listed with pos (0,0)
      (sigma-group-put-window-in-group w index)
      (sigma-group-delete-window-from-group w index2)
      (sigma-group-refresh-group w 'move)
      (call-window-hook 'sigma-group-grouped-window-hook w)
      ))

  ;; Release active window from its group
  (define (sigma-group-release-window w)
    (interactive "%W")
    (sigma-group-delete-window-from-groups w)
    (call-window-hook 'sigma-group-released-window-hook w)
    (sigma-group-make-new-group w)
    (sigma-group-refresh-group w 'frame)
    )

  ;; Return the window at position (pos+n) in window's group
  (define (sigma-group-offset win n)
    (let* ((gr (siggroup-window-list (sigma-group-find-window win)))
           (size (length gr))
           (r (sigma-group-rank win gr)))
      (nth (modulo (+ r n) size) gr)))

  ;; Predicate : true <=> w1 and w2 are grouped together
  (define (sigma-group-same-group-p w1 w2)
    (let ((id1 (window-get w1 'siggroup))
          (id2 (window-get w2 'siggroup)))
    (and (equal id1 id2)
         (not (null id1)))))

  ;; Raise left window in current group
  (define (sigma-group-raise-left-window w)
    (interactive "%W")
    (let ((win (sigma-group-offset w -1)))
      (raise-window win)
      (set-input-focus win))
    )

  ;; Raise right window in current group
  (define (sigma-group-raise-right-window w)
    (interactive "%W")
    (let ((win (sigma-group-offset w 1)))
      (raise-window win)
      (set-input-focus win))
    )

  ;; Map a function over the other windows in win's group
  (define (sigma-group-map-other-grouped-windows win func)
    (mapcar func
            (delete-if
             (lambda (w) (eq w win))
             (siggroup-window-list (sigma-group-find-window win)))))

  ;;;;;;;;;;;;;;;;;;;
  ;; Hooking stuff :
  ;;

  (add-hook 'window-state-change-hook
            (lambda (win args)
              (if (eq 'sticky args)
                  (sigma-group-refresh-group win 'sticky))))
  (add-hook 'while-moving-hook (lambda (win) (sigma-group-refresh-group win 'move)))
  (add-hook 'while-resizing-hook (lambda (win) (sigma-group-refresh-group win 'resize)))
  (add-hook 'window-maximized-hook (lambda (win args) (sigma-group-refresh-group win 'resize)))
  (add-hook 'window-unmaximized-hook (lambda (win args) (sigma-group-refresh-group win 'resize)))
  (add-hook 'after-move-hook (lambda (win args) (sigma-group-refresh-group win 'move)))
  (add-hook 'after-resize-hook (lambda (win args) (sigma-group-refresh-group win 'resize)))
  (add-hook 'shade-window-hook (lambda (win) (sigma-group-refresh-group win 'shade)))
  (add-hook 'unshade-window-hook (lambda (win) (sigma-group-refresh-group win 'unshade)))
  (add-hook 'destroy-notify-hook sigma-group-delete-window-from-groups)
  ;; (add-hook 'add-to-workspace-hook (lambda (win lws)
  ;;                                    (if sigma-group-move-lock
  ;;                                        (progn
  ;;                                          (setq sigma-group-move-lock nil)
  ;;                                          (mapcar (lambda (w) (copy-window-to-workspace w lws))
  ;;                                                  (delete-if
  ;;                                                   (lambda (w) (eq w win))
  ;;                                                   (siggroup-window-list (sigma-group-find-window win))))
  ;;                                          (setq sigma-group-move-lock t)
  ;;                                        ))))
  ;; (add-hook 'remove-from-workspace-hook (lambda (win lws)
  ;;                                    (if sigma-group-move-lock
  ;;                                        (progn
  ;;                                          (setq sigma-group-move-lock nil)
  ;;                                          (mapcar (lambda (w) (window-remove-from-workspace w lws))
  ;;                                                  (delete-if
  ;;                                                   (lambda (w) (eq w win))
  ;;                                                   (siggroup-window-list (sigma-group-find-window win))))
  ;;                                          (setq sigma-group-move-lock t)
  ;;                                        ))))

)
